''
'' Quite Okay Image Viewer (FreeBasic)
'' by ssjx (https://ssjx.co.uk)
''
''
'' The QOI format was created by Dominic Szablewski,
'' more information can be found here https://qoiformat.org/
''
''
'' Compile with:
''		fbc *.bas
''
'' For additonal checks/safety, compile with: 
''		fbc -exx *.bas
''
''
'' Version 0.3 (12/07/2024)
'' + Rearranged decoder and fixed stray line on demo dice image
''
'' Version 0.2 (03/01/2022)
'' + Resolution check
'' + Added screenlocks (forgot these...)
''
'' Version 0.1 (02/01/2022)
'' + First release

const TITLE="QOI Viewer"
const VERSION="v0.3 (12/07/2024)"
const SITEURL="https://ssjx.co.uk"

'' Main Header 14 bytes
type my_qoi_header Field = 1
	magic as zstring *4 	''QOIF
	w as ulong
	h as ulong
	channels as ubyte		' 3=RGB, 4=RGBA
	colorspace as ubyte
end type

type my_rgba Field = 1
	r as ubyte
	g as ubyte
	b as ubyte
	a as ubyte
end type

type my_rgb Field = 1
	r as ubyte
	g as ubyte
	b as ubyte
end type

dim shared cache(63) as my_rgba  	'' Remember FB array is inclusive!

dim shared pix_rgba as my_rgba  	'' 
dim shared pix_rgb as my_rgb  		'' 

dim shared qoi_header as my_qoi_header


sub cache_add(r as uinteger, g as uinteger, b as uinteger, a as uinteger)
	dim as integer idx=((r*3)+(g*5)+(b*7)+(a*11)) mod 64
	
	cache(idx).r=r
	cache(idx).g=g
	cache(idx).b=b
	cache(idx).a=a
end sub

''
''
''

function swaplong(in as ulong) as uinteger
	dim as ulong value
	value =((in and &hff000000) shr 24) + ((in and &h00ff0000) shr 8)+((in and &hff) shl 24)+((in and &hff00) shl 8)
	return value
end function

''
''
''

dim as string filename=""
filename = Command(1)

print TITLE+" "+VERSION+" by ssjx - "+SITEURL
	
if filename="" then
	print "Usage : qoi myfile.qoi"
	end 
else
	print "Loading file "+filename
end if
''

if  Open(filename For binary access read As #1)=0 then
	print "Loading..."
	get #1,,qoi_header
	
	dim as string magic=left(qoi_header.magic,4)
			
	print "--> Magic  : "+magic
	
	if magic<>"qoif" then
		print "This does not look like a valid QOI file..."
		end
	end if
	
	
	dim as integer actual_w=swaplong(qoi_header.w)
	dim as integer actual_h=swaplong(qoi_header.h)
	
	print "--> Dimensions  : "+str(actual_w)+" x "+str(actual_h)
	
	if actual_w<=0 or actual_h<=0 then
		print "These dimensions do not seem right..."
		end
	end if
	
	print "--> Channels  : "+str(qoi_header.channels)
	print "--> Colorspace  : "+str(qoi_header.colorspace)
	
	''
	dim as my_rgba pix(actual_w*actual_h)
	
	dim as ubyte tag
	dim as integer ps=0
	
	pix_rgba.r=0
	pix_rgba.g=0
	pix_rgba.b=0
	pix_rgba.a=255
				
	do
		dim as ubyte repeat=0
		dim as boolean found_tag=false
		
		get #1,,tag
		
		select case tag
			case 254
				'' RGB
				get #1,,pix_rgb
				pix_rgba.r=pix_rgb.r
				pix_rgba.g=pix_rgb.g
				pix_rgba.b=pix_rgb.b
				found_tag=true
			case 255
				'' RGBA
				get #1,,pix_rgba
				found_tag=true
		end select
		 
		if found_tag=false then
			'' Must be a 2 Bit Tag (1100 0000)
			dim as ubyte col=tag
			
			tag=(tag shr 6) and &h03
			
			select case tag
			
				case 0
					'' From pixel cache
					dim as integer color_idx=(col and &h3F)
				
					'' WORKING?
					if color_idx>=0 and color_idx<=63 then 
						pix_rgba.r=cache(color_idx).r
						pix_rgba.g=cache(color_idx).g
						pix_rgba.b=cache(color_idx).b
						pix_rgba.a=cache(color_idx).a
					end if
				
				case 1
					'' Difference
					dim as ubyte dr=(col shr 4) and &h03
					dim as ubyte dg=(col shr 2) and &h03
					dim as ubyte db=(col and &h03) 
			
					dr-=2
					dg-=2
					db-=2
					
					pix_rgba.r+=dr
					pix_rgba.g+=dg
					pix_rgba.b+=db
					
				case 2
					'' Luma
					dim as ubyte luma
					get #1,,luma
					
					dim as integer luma_g=(col and &h3f)-32
					
					dim as integer luma_r=((luma and &hf0) shr 4)+luma_g-8
					dim as integer luma_b=(luma and &h0f)+luma_g-8
					
					pix_rgba.r+=luma_r
					pix_rgba.g+=luma_g
					pix_rgba.b+=luma_b
					
				case 3
				
					'' Run
					repeat=(col and &h3F)
			end select
			
		end if
	
		if repeat>0 then
			for r as integer=0 to repeat
				pix(ps).r=pix_rgba.r
				pix(ps).g=pix_rgba.g
				pix(ps).b=pix_rgba.b
				pix(ps).a=pix_rgba.a
				ps+=1
			next 
		else
			pix(ps).r=pix_rgba.r
			pix(ps).g=pix_rgba.g
			pix(ps).b=pix_rgba.b		
			ps+=1
			
			cache_add(pix_rgba.r, pix_rgba.g, pix_rgba.b, pix_rgba.a)
		end if
		
	loop until (ps>=(actual_w*actual_h))
	close #1

	print "Displaying!"
	
	screenres actual_w,actual_h,32
	
	windowtitle TITLE+" "+VERSION+" - "+SITEURL+" - Press Space to Exit" 
	
	ps=0
	screenlock
		for y as integer=0 to actual_h-1
			for x as integer=0 to actual_w-1
				pset (x,y),rgb(pix(ps).r,pix(ps).g,pix(ps).b)
				ps+=1
			next x
		next y
	screenunlock
	sleep

else
	print "Could not open image..."
end if


		
		
